Imports System.Runtime.InteropServices ' The Marshal class is derieved from here
Imports System.Drawing.Imaging         ' BitmapData structure from here

' Based on Image Processing for Dummies with C# and GDI+ Part 1 Article
' URL: http://www.codeproject.com/csharp/csharpgraphicfilters11.asp

' Convert to VB.NET by Max Raskin ( max2k@netvision.net.il )

Public Class Filters

    Shared bmData As BitmapData
    Shared ptr As System.IntPtr
    Shared Red As Integer, Green As Integer, Blue As Integer
    Shared x As Integer, y As Integer
    Shared nOffset As Integer

    ' All of the functions below are very simillar in structure.
    ' All functions lock the bits of the bitmap to gain access to the memory
    ' and all functions contain 2 loops one within the other which modify the pixels.

    Public Shared Function ContrastFilter(ByVal ContrastValue As Integer, ByRef b As Bitmap) As Bitmap
        ' No support for 256 colors...
        If b.PixelFormat = PixelFormat.Format8bppIndexed Then
            MsgBox("256 colors bitmap are not supported.", MsgBoxStyle.Critical Or MsgBoxStyle.ApplicationModal, "Error")
            Return Nothing
        End If
        If (ContrastValue < -100 Or ContrastValue > 100) Then Return Nothing
        ' The LockBits locks the bitmap and let's us modify its bytes in the memory
        bmData = b.LockBits(New Rectangle(0, 0, b.Width, b.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb)
        ' This is the pointer to the first pixel of the bitmap
        ptr = bmData.Scan0
        nOffset = bmData.Stride - b.Width * 3
        Dim pixel As Double
        Dim contrast As Double = (100 + ContrastValue) / 100
        contrast *= contrast
        ' Loop through rows of the bitmap an modify contrast per pixel according to the formula
        For y = 0 To b.Height - 1
            For x = 0 To b.Width - 1
                ' note that colors are read BGR and not RGB since they're written that way.
                ' The Marshal class let's us read and write bytes in arrays through pointers,
                Blue = Marshal.ReadByte(ptr, 0)
                pixel = Blue / 255
                pixel -= 0.5
                pixel *= contrast
                pixel += 0.5
                pixel *= 255
                If (pixel < 0) Then pixel = 0
                If (pixel > 255) Then pixel = 255
                Marshal.WriteByte(ptr, 0, CByte(pixel))

                Green = Marshal.ReadByte(ptr, 1)
                pixel = Green / 255
                pixel -= 0.5
                pixel *= contrast
                pixel += 0.5
                pixel *= 255
                If (pixel < 0) Then pixel = 0
                If (pixel > 255) Then pixel = 255
                Marshal.WriteByte(ptr, 1, CByte(pixel))

                Red = Marshal.ReadByte(ptr, 2)
                pixel = Red / 255
                pixel -= 0.5
                pixel *= contrast
                pixel += 0.5
                pixel *= 255
                If (pixel < 0) Then pixel = 0
                If (pixel > 255) Then pixel = 255
                Marshal.WriteByte(ptr, 2, CByte(pixel))

                ' In C++ we would simply do ptr += 3; but here it VB.NET
                ' it gets ugly, but it works (op_Explicit converts from Int32 to an IntPtr
                ' pointer).
                ' We increase the pointer by 3 (3 bytes for BGR):
                ptr = IntPtr.op_Explicit(ptr.ToInt32 + 3)
            Next
            ' Advance an offset
            ptr = IntPtr.op_Explicit(ptr.ToInt32 + nOffset)
        Next
        b.UnlockBits(bmData)
        Return b
    End Function

    Public Shared Function BrightnessFilter(ByVal BrightnessValue As Integer, ByRef b As Bitmap) As Bitmap
        '===================================================================================='
        ' This function is quite simple, it adds/subtracts the BrightnessValue to each byte  '
        ' and so it increases/decreases the brightness of the pixels.                        '
        '===================================================================================='
        If b.PixelFormat = PixelFormat.Format8bppIndexed Then
            MsgBox("256 colors bitmap are not supported.", MsgBoxStyle.Critical Or MsgBoxStyle.ApplicationModal, "Error")
            Return Nothing
        End If
        If BrightnessValue = 0 Then Return Nothing
        bmData = b.LockBits(New Rectangle(0, 0, b.Width, b.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb)
        ptr = bmData.Scan0
        nOffset = bmData.Stride - b.Width * 3
        For y = 0 To b.Height - 1
            For x = 0 To (b.Width * 3) - 1
                Dim bByte As Integer = Marshal.ReadByte(ptr, 0)
                bByte += BrightnessValue
                ' Make sure that the byte doesn't go beyond 255 or beneath 0
                If bByte > 255 Then bByte = 255
                If bByte < 0 Then bByte = 0
                ' Write the new byte and loop until all bytes are modified
                Marshal.WriteByte(ptr, 0, CByte(bByte))
                ' In C++ we would simply do ptr += 3; but here it VB.NET
                ' it gets ugly, but it works (op_Explicit converts from Int32 to an IntPtr
                ' pointer).
                ' We increase the pointer by 3 (3 bytes for BGR):
                ptr = IntPtr.op_Explicit(ptr.ToInt32 + 1)
            Next
            ' Advance an offset
            ptr = IntPtr.op_Explicit(ptr.ToInt32 + nOffset)
        Next
        b.UnlockBits(bmData)
        Return b
    End Function

    Public Shared Function Invert(ByRef b As Bitmap) As Bitmap
        ' ========================================================================'
        ' This function simply loops through all bytes of the image and           '
        ' subtracts from 255 and so we get the opposite RGB values of each pixel. '
        ' ========================================================================'
        If b.PixelFormat = PixelFormat.Format8bppIndexed Then
            MsgBox("256 colors bitmap are not supported.", MsgBoxStyle.Critical Or MsgBoxStyle.ApplicationModal, "Error")
            Return Nothing
        End If

        bmData = b.LockBits(New Rectangle(0, 0, b.Width, b.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb)
        ptr = bmData.Scan0
        nOffset = bmData.Stride - b.Width * 3
        For y = 0 To b.Height - 1
            ' b.Width * 3 because we need to go through all RGB values (BGR in this case but it doesn't matter now.)
            For x = 0 To (b.Width * 3) - 1
                Marshal.WriteByte(ptr, 0, CByte(255 - Marshal.ReadByte(ptr, 0)))
                ' In C++ we would simply do ptr += 3; but here it VB.NET
                ' it gets ugly, but it works (op_Explicit converts from Int32 to an IntPtr
                ' pointer).
                ' Advance one pixel:
                ptr = IntPtr.op_Explicit(ptr.ToInt32 + 1)
            Next
            ' Advance an offset:
            ptr = IntPtr.op_Explicit(ptr.ToInt32 + nOffset)
        Next
        b.UnlockBits(bmData)
        Return b
    End Function

    Public Shared Function Grayscale(ByRef b As Bitmap) As Bitmap
        ' ============================================================'
        ' This function uses a formula to convert pixels to grayscale '
        ' ============================================================'
        If b.PixelFormat = PixelFormat.Format8bppIndexed Then
            MsgBox("256 colors bitmap are not supported.", MsgBoxStyle.Critical Or MsgBoxStyle.ApplicationModal, "Error")
            Return Nothing
        End If
        bmData = b.LockBits(New Rectangle(0, 0, b.Width, b.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb)
        ptr = bmData.Scan0
        nOffset = bmData.Stride - b.Width * 3
        Dim bVal As Byte
        ' Loop throughout the bitmap and change each pixel according to the formula.
        For y = 0 To b.Height - 1
            For x = 0 To b.Width - 1
                Blue = Marshal.ReadByte(ptr, 0)
                Green = Marshal.ReadByte(ptr, 1)
                Red = Marshal.ReadByte(ptr, 2)
                bVal = CByte(0.299 * Red + 0.587 * Green + 0.114 * Blue)
                Marshal.WriteByte(ptr, 0, bVal)
                Marshal.WriteByte(ptr, 1, bVal)
                Marshal.WriteByte(ptr, 2, bVal)
                ptr = IntPtr.op_Explicit(ptr.ToInt32 + 3)
            Next
            ptr = IntPtr.op_Explicit(ptr.ToInt32 + nOffset)
        Next
        b.UnlockBits(bmData)
        Return b
    End Function

    Public Shared Function AdjustColors(ByRef b As Bitmap, ByVal RedValue As Integer, ByVal BlueValue As Integer, ByVal GreenValue As Integer) As Bitmap
        If b.PixelFormat = PixelFormat.Format8bppIndexed Then
            MsgBox("256 colors bitmap are not supported.", MsgBoxStyle.Critical Or MsgBoxStyle.ApplicationModal, "Error")
            Return Nothing
        End If
        ' This function simply adds/subtracts RGB values
        ' You can remove a specific color by typing -255        
        bmData = b.LockBits(New Rectangle(0, 0, b.Width, b.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb)
        ptr = bmData.Scan0
        Dim bVal As Byte
        nOffset = bmData.Stride - b.Width * 3
        For y = 0 To b.Height - 1
            For x = 0 To b.Width - 1
                Blue = Marshal.ReadByte(ptr, 0)
                Green = Marshal.ReadByte(ptr, 1)
                Red = Marshal.ReadByte(ptr, 2)

                ' Red
                Red += RedValue
                Red = Math.Max(Red, 0)
                Red = Math.Min(Red, 255)

                ' Green
                Green += GreenValue
                Green = Math.Max(Green, 0)
                Green = Math.Min(Green, 255)

                ' Blue
                Blue += BlueValue
                Blue = Math.Max(Blue, 0)
                Blue = Math.Min(Blue, 255)

                ' BGR Format...
                Marshal.WriteByte(ptr, 0, CByte(Blue))
                Marshal.WriteByte(ptr, 1, CByte(Green))
                Marshal.WriteByte(ptr, 2, CByte(Red))
                ptr = IntPtr.op_Explicit(ptr.ToInt32 + 3)
            Next
            ptr = IntPtr.op_Explicit(ptr.ToInt32 + nOffset)
        Next
        b.UnlockBits(bmData)
        Return b
    End Function

    Public Shared Function AdjustGamma(ByRef b As Bitmap, ByVal RedValue As Double, ByVal GreenValue As Double, ByVal BlueValue As Double) As Bitmap

        ' I know there is a gamma function in GDI+ for images but why not see how its done ?

        If b.PixelFormat = PixelFormat.Format8bppIndexed Then
            MsgBox("256 colors bitmap are not supported.", MsgBoxStyle.Critical Or MsgBoxStyle.ApplicationModal, "Error")
            Return Nothing
        End If


        If (RedValue < 0.2 Or RedValue > 5) Then Return Nothing
        If (GreenValue < 0.2 Or GreenValue > 5) Then Return Nothing
        If (BlueValue < 0.2 Or BlueValue > 5) Then Return Nothing

        Dim redGamma(256) As Byte
        Dim greenGamma(256) As Byte
        Dim blueGamma(256) As Byte

        Dim i As Integer

        For i = 0 To 255
            redGamma(i) = CByte(Math.Min(255, CInt(((255.0 * Math.Pow(i / 255.0, 1.0 / RedValue)) + 0.5))))
            greenGamma(i) = CByte(Math.Min(255, CInt(((255.0 * Math.Pow(i / 255.0, 1.0 / GreenValue)) + 0.5))))
            blueGamma(i) = CByte(Math.Min(255, CInt(((255.0 * Math.Pow(i / 255.0, 1.0 / BlueValue)) + 0.5))))
        Next

        bmData = b.LockBits(New Rectangle(0, 0, b.Width, b.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb)
        ptr = bmData.Scan0
        nOffset = bmData.Stride - b.Width * 3
        For y = 0 To b.Height - 1
            For x = 0 To b.Width - 1
                Marshal.WriteByte(ptr, 0, blueGamma(Marshal.ReadByte(ptr, 0)))
                Marshal.WriteByte(ptr, 1, greenGamma(Marshal.ReadByte(ptr, 1)))
                Marshal.WriteByte(ptr, 2, redGamma(Marshal.ReadByte(ptr, 2)))
                ptr = IntPtr.op_Explicit(ptr.ToInt32 + 3)
            Next
            ptr = IntPtr.op_Explicit(ptr.ToInt32 + nOffset)
        Next
        b.UnlockBits(bmData)
        Return b
    End Function

End Class
